/*
Written based on explanation in "Introduction to algorithms" by Thomas H. Cormen et al.
This algorithm only works with signals of length equal to power of 2.
*/
#include <math.h>
#include <stdio.h>

// define TAU as 2*PI
// 1 tau = 2*pi radians = full circle turn
// This constant is defined for floats. If you wish to use doubles, add more digits
#define TAU 6.283185307

// complex number structure
typedef struct cnum
{
    float r; // real part
    float i; // imaginary part
} cnum;

// add complex numbers
cnum cnum_add(cnum a, cnum b)
{
    cnum out;
    out.r = a.r + b.r;
    out.i = a.i + b.i;
    return out;
}

// substract complex numbers
cnum cnum_sub(cnum a, cnum b)
{
    cnum out;
    out.r = a.r - b.r;
    out.i = a.i - b.i;
    return out;
}

// multiply complex numbers
// (ar + i * ai) * (br + i* bi) = (ar*br - ai*bi) + i (ar*bi + ai*br)
cnum cnum_mul(cnum a, cnum b)
{
    cnum out;
    out.r = a.r*b.r - a.i*b.i;
    out.i = a.r*b.i + a.i*b.r;
    return out;
}

/*
Reverse last bits from value.

Params:
    value - value whose bits are to be reversed
    bits - how many last bits from value are to be reversed
Return:
    value with reversed last bits

Example of bit reversal for bits=4:
    0000 -> 0000
    1000 -> 0001
    1101 -> 1011
    1111 -> 1111
*/
int reverse_bits(int value, int bits)
{
    int out = 0;
    int bit;

    for(int i=0; i<bits; ++i)
    {
        // get value of last bit
        bit = value & 1;
        // append bit to out
        out = (out << 1) | bit;
        // remove last bit from value
        value = value >> 1;
    }

    return out;
}

void copy_with_bitreverse_positions(cnum *in, cnum *out, int n, int bits)
{
    for(int i=0; i<n; ++i)
    {
        int new_pos = reverse_bits(i, bits);
        out[new_pos] = in[i];
    }
}

/*
Compute Fast Fourier Transform.

Take values from in, compute fft and save result in out.

Params:
    in - samples of signal in time domain
    out - amplitudes of frequencies in frequency domain
    n - length of in, out
        n must be power of 2
*/
void fft(cnum *in, cnum *out, int n)
{
    int bits = log2f(n);

    // rearrange input in a way that we can avoid recurrence
    copy_with_bitreverse_positions(in, out, n, bits);

    /*
    Level 0 in fft recurrence tree is at leaves.
    Leaf has a single value, and dft(single_value) = single_value, so there is no
    need to do any compuations for that level.
    Next we need to merge leaves into 2-point dfts. Then these into 4-point dfts, these into 8-point dfts
    etc.

    Level 0 contains n 1-point dfts.
    Level 1 contains n/2 2-points dfts.
    Level 2 contains n/4 4-points dfts.
    Level 3 contains n/8 8-point dfts.
    etc.

    So level contains 2^level points in single dft.
    */

    int max_level = bits + 1;
    for(int level=1; level<max_level; ++level)
    {
        int points_in_dft = 1<<level; // 2^level
        float points_in_dft_float = (float) points_in_dft;

        // merge sub-dfts in current level
        int dfts_in_level = n / points_in_dft;
        for(int dft_index=0; dft_index<dfts_in_level; ++dft_index)
        {
            int dft_start_offset = dft_index * points_in_dft;

            // e^(i*2pi*0) = cos(0) + i*sin(0) = 1 + i*0;
            cnum omega;
            omega.r = 1.0f;
            omega.i = 0.0f;

            // e^-(i*2pi/points_in_dft) = cos(2pi/points_in_dft) - i*sin(cos(2pi/points_in_dft))
            cnum omega_n;
            omega_n.r = cosf(TAU/points_in_dft_float);
            omega_n.i = -sinf(TAU/points_in_dft_float); // notice minus

            // for points in current dft
            // k travels only half of the points because we are using even and odd points positions
            // iterative fft avoids dividing points into separate even and odd arrays, but keeps them in single array
            for(int k=0; k<points_in_dft/2; ++k)
            {
                cnum even = out[dft_start_offset + k]; // even[k]
                cnum odd = out[dft_start_offset + k + points_in_dft/2]; // odd[k]

                odd = cnum_mul(omega, odd);

                out[dft_start_offset + k] = cnum_add(even, odd);
                out[dft_start_offset + k + points_in_dft/2] = cnum_sub(even, odd);

                omega = cnum_mul(omega, omega_n);
            }
        }
    }
}

/*
Computer inverse Fast Fourier Transform.

Take values from in, compute ifft and save result in out.

Params:
    in - amplitudes of frequencies in frequency domain
    out - samples of signal in time domain
    n - length of in, out
        n must be power of 2

Implementation notice:
Inverse FFT is mostly the same as forward FFT. Look at these equations:
a) y(k) = sum(n=0, N-1)( x(n) * e^(-j*2pi*k/N*n) )
b) x(n) = 1/N * sum(k=0, N-1)( y(k) * e^(j*2pi*k/N*n) )

Equation a) describes how to compute k-th value in forward DFT.
Equation b) describes how to compute n-th value in inverse DFT.

There are only three differences:
1. use y(k) intead of x(n)
2. do not use minus in exponent e^-(j*2pi*k/N*n)
3. divide output by number of samples N

These are very easy to incorporate into fft to get inverse fft.
*/
void ifft(cnum *in, cnum *out, int n)
{
    int bits = log2f(n);

    copy_with_bitreverse_positions(in, out, n, bits);

    int max_level = bits + 1;
    for(int level=1; level<max_level; ++level)
    {
        int points_in_dft = 1<<level;
        float points_in_dft_float = (float) points_in_dft;

        int dfts_in_level = n / points_in_dft;
        for(int dft_index=0; dft_index<dfts_in_level; ++dft_index)
        {
            int dft_start_offset = dft_index * points_in_dft;

            cnum omega;
            omega.r = 1.0f;
            omega.i = 0.0f;

            cnum omega_n;
            omega_n.r = cosf(TAU/points_in_dft_float);
            omega_n.i = sinf(TAU/points_in_dft_float); // no minus in inverse fft

            for(int k=0; k<points_in_dft/2; ++k)
            {
                cnum even = out[dft_start_offset + k];
                cnum odd = out[dft_start_offset + k + points_in_dft/2];

                odd = cnum_mul(omega, odd);

                out[dft_start_offset + k] = cnum_add(even, odd);
                out[dft_start_offset + k + points_in_dft/2] = cnum_sub(even, odd);

                omega = cnum_mul(omega, omega_n);
            }
        }
    }

    // divide values by n
    for(int i=0; i<n; ++i)
    {
        out[i].r /= n;
        out[i].i /= n;
    }
}

#define samples 4
cnum in[samples];
cnum out[samples];
cnum in2[samples];

int main()
{
    int n = samples;

    in[0].r = 1.0;
    in[0].i = 2.0;

    in[1].r = 3.0;
    in[1].i = 4.0;

    in[2].r = 2.0;
    in[2].i = 4.0;

    in[3].r = 3.0;
    in[3].i = 7.0;

    printf("in:\n");
    for(int i=0; i<n; ++i)
    {
        printf("%f%+f\n", in[i].r, in[i].i);
    }

    fft(in, out, n);

    printf("out:\n");
    for(int i=0; i<n; ++i)
    {
        printf("%f%+f\n", out[i].r, out[i].i);
    }

    ifft(out, in2, n);

    printf("reconstructed in:\n");
    for(int i=0; i<n; ++i)
    {
        printf("%f%+f\n", in2[i].r, in2[i].i);
    }


    return 0;
}
